import fcntl
import os
import pty
import select
import signal
import subprocess
from threading import Thread

from .base import ProcessNode


class ShellProcessIO(ProcessNode):
    """
    The widget will start a bash shell in the background and use a pseudo TTY to control it.  It then
    starts a thread to transfer any data between the two processes (the one running this widget and
    the bash shell).
    """

    def __init__(self, name="Shell", command="bash -i"):
        super().__init__(name=name, command=command)

        # Open a pseudo TTY to control the interactive session.
        # Make it non-blocking.
        self._master, self._slave = pty.openpty()
        fl = fcntl.fcntl(self._master, fcntl.F_GETFL)
        fcntl.fcntl(self._master, fcntl.F_SETFL, fl | os.O_NONBLOCK)

        self._process = subprocess.Popen(
            self.command.split(" "),
            preexec_fn=os.setsid,
            stdin=self._slave,
            stdout=self._slave,
            stderr=self._slave,
        )

        # Start the shell and thread to pull data from it.
        self._running = True
        self._extraction_thread = Thread(
            name=f"{self.name} Extractor Thread", target=self._background, daemon=True
        )
        self._extraction_thread.start()

    def _background(self):
        """
        Background thread running the IO between the widget and the TTY session.
        """
        while self._running:
            # Wait for any of the file descriptors to have data, with a 1 second timeout
            ready, _, _ = select.select([self._master], [], [], 0.1)
            self._onboard_new_subscribers()

            for stream in ready:
                data_bytes = b""
                while self._running:
                    try:
                        new_bytes = os.read(stream, self._READ_MAX_BYTES)
                    except BlockingIOError:
                        break
                    data_bytes += new_bytes

                self._record_and_publish(data_bytes)

    def write(self, val: bytes):
        """Write a character to the tty"""
        os.write(self._master, val)

    def send_sigint(self):
        """Simulate the user sending Ctrl+C to the underlying shell"""
        self.write(b"\x03")
        # self._process.send_signal(signal.SIGINT)

    def close(self, timeout=None):
        """While not necessary, sending a sigint significantly speeds up the closing
        process, which is great for tests."""
        self.send_sigint()
        super().close(timeout=timeout)



from pathlib import Path

from asciimatics import effects, renderers, widgets
from asciimatics.exceptions import InvalidFields
from asciimatics.widgets import Frame, Layout

from groklog.filter_manager import DuplicateFilterError, FilterManager

from . import scene_names
from .base_app import BaseApp


class FilterCreator(BaseApp):
    _NEW_FILTER_PARAMETERS_COLUMN = 1
    _SYSTEM_SETTINGS_COLUMN = 3
    _SAVE_PATH_LABEL = "save_path"

    def __init__(self, screen, filter_manager: FilterManager, profile_path: Path):
        super().__init__(
            screen,
            screen.height,
            screen.width,
            has_border=True,
            can_scroll=True,
            name="Benis",
            title="Create Filter",
        )
        self._filter_manager: FilterManager = filter_manager
        self.profile_path: Path = profile_path

        dialog_layout = Layout([1, 30, 1, 30, 1], fill_frame=True)

        self.add_layout(dialog_layout)

        # Add the widgets for the left column
        dialog_layout.add_widget(
            widgets.Label("System Configuration", align="^", height=2),
            column=self._SYSTEM_SETTINGS_COLUMN,
        )
        dialog_layout.add_widget(
            widgets.Text(
                label="Save Path",
                name=self._SAVE_PATH_LABEL,
                on_change=None,
                validator=None,
            ),
            column=self._SYSTEM_SETTINGS_COLUMN,
        )

        # Add the widgets for the right column
        dialog_layout.add_widget(
            widgets.Label("New Filter Parameters", align="^", height=2),
            column=self._NEW_FILTER_PARAMETERS_COLUMN,
        )
        self.source_drop_down = widgets.DropdownList(
            options=[],
            label="Input Source",
            name=FilterManager.FILTER_PARENT,
        )

        dialog_layout.add_widget(
            self.source_drop_down, column=self._NEW_FILTER_PARAMETERS_COLUMN
        )
        dialog_layout.add_widget(
            widgets.Text(
                label="Filter Name",
                name=FilterManager.FILTER_NAME,
                validator=lambda val: 0 < len(val) < 15,
            ),
            column=self._NEW_FILTER_PARAMETERS_COLUMN,
        )
        dialog_layout.add_widget(
            widgets.Text(
                label="Command",
                name=FilterManager.FILTER_COMMAND,
                validator=lambda val: len(val) > 0,
            ),
            column=self._NEW_FILTER_PARAMETERS_COLUMN,
        )

        # I put two dividers here, one invisible. The invisible one pushes the
        # visible one all the way to the bottom, otherwise, it'll appear in the
        # middle.
        for column in range(len(dialog_layout._columns)):
            dialog_layout.add_widget(
                widgets.Divider(height=widgets.Widget.FILL_FRAME, draw_line=False),
                column=column,
            )
            dialog_layout.add_widget(widgets.Divider(), column=column)

        # Create the Tab Layout and buttons for it
        save_filter_button = widgets.Button("Save", self.save_filters)
        cancel_filter_button = widgets.Button("Cancel", self.cancel_contents)
        save_cancel_layout = Layout([7, 1, 1])
        self.add_layout(save_cancel_layout)
        save_cancel_layout.add_widget(save_filter_button, 1)
        save_cancel_layout.add_widget(cancel_filter_button, 2)

        self.fix()

    def save_filters(self):
        # TOOD: Add logic to save things here
        try:
            self.save(validate=True)
        except InvalidFields:
            self.display_popup("Please fix any fields highlighted in yellow.", ["Ok"])
            return

        try:
            parent_index = self.data[FilterManager.FILTER_PARENT]
            parent_name = self.source_drop_down.options[parent_index][0]
            parent_filter = self._filter_manager.get_filter(parent_name)
            self._filter_manager.create_filter(
                name=self.data[FilterManager.FILTER_NAME],
                command=self.data[FilterManager.FILTER_COMMAND],
                parent=parent_filter,
            )
        except DuplicateFilterError:
            self.display_popup("There already exists a filter with that name!", ["Ok"])
            return
        self.profile_path = Path(self.data[self._SAVE_PATH_LABEL])

        self._filter_manager.save_profile(self.profile_path)

        self.reset()
        self.change_scene(scene_name=scene_names.SHELL_VIEW)

    def reset(self):
        self.source_drop_down.options = [
            (f.name, i) for i, f in enumerate(self._filter_manager)
        ]
        super().reset()
        self.data = {
            FilterManager.FILTER_PARENT: 0,
            FilterManager.FILTER_NAME: "",
            FilterManager.FILTER_COMMAND: "",
            self._SAVE_PATH_LABEL: str(self.profile_path),
        }

    def cancel_contents(self):
        self.reset()
        self.change_scene(scene_name=scene_names.SHELL_VIEW)



from typing import Callable, List

from asciimatics import effects, renderers, widgets
from asciimatics.event import KeyboardEvent
from asciimatics.exceptions import NextScene, StopApplication
from asciimatics.screen import Screen

from groklog.ui.theming import set_theme


class BaseApp(widgets.Frame):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        set_theme(self)

    def change_scene(self, scene_name: str):
        raise NextScene(scene_name)

    def display_toast(self, message):
        """Display a temporary message for a few seconds near the bottom of
        the screen."""
        self._scene.add_effect(
            effects.Print(
                self.screen,
                renderers.SpeechBubble(text=message, uni=True),
                speed=1,
                transparent=False,
                clear=True,
                delete_count=30,
                y=int(self.screen.height * 0.8),
            ),
        )

    def display_popup(
        self,
        text: str,
        buttons: List[
            str,
        ],
        callback: Callable[[str], None] = None,
        theme="warning",
    ):
        self._scene.add_effect(
            widgets.PopUpDialog(
                self._screen,
                text,
                buttons,
                has_shadow=True,
                on_close=callback,
                theme=theme,
            )
        )

    def process_event(self, event):
        """Handle shutting down the program if escape is pressed"""
        if isinstance(event, KeyboardEvent):
            if event.key_code == Screen.KEY_ESCAPE:
                # Close the program when Escape or End is pressed
                raise StopApplication("")
        return super().process_event(event)



import copy
from queue import Queue

from asciimatics.parsers import AnsiTerminalParser
from asciimatics.strings import ColouredText
from asciimatics.widgets import TextBox

from groklog.process_node import GenericProcessIO, ProcessNode

_line_cache = {}


class FilterViewer(TextBox):
    def __init__(self, filter: GenericProcessIO, height: int):
        super().__init__(
            height,
            name=f"FilterViewer-{filter.name}-{filter.command})",
            readonly=True,
            as_string=False,
            line_wrap=True,
            parser=AnsiTerminalParser(),
        )
        self.filter = filter
        self._data_queue = Queue()
        self.custom_colour = "filter_viewer"
        self._value = [ColouredText("", self._parser, colour=None)]

        # Create subscriptions
        self._processed_data_queue = Queue()
        """self._add_stream pushes to here, and self.update pulls the results"""

        filter.subscribe_with_history(
            ProcessNode.Topic.STRING_DATA_STREAM, self._add_stream, blocking=False
        )

    def process_event(self, event):
        # TODO: Change keyboard up/down events to immediately move up/down a line.
        return super().process_event(event)

    def update(self, frame_no):
        changed = False
        while self._processed_data_queue.qsize():
            changed = True
            self._value += self._processed_data_queue.get_nowait()

        if changed:
            self.reset()

        return super().update(frame_no)

    def _add_stream(self, append_logs: str):
        """Append text to the log stream. This function should receive input from
        the filter and display it."""

        # TODO: Consider limiting split size on massive files?
        # TODO: Consider appending only after parsing colors

        last_colour = self._value[-1].last_colour

        # TODO: remove multiple newlines
        new_lines = []
        for line in append_logs.split("\n"):
            # TODO
            cache_key = (line, last_colour, self.filter.name)
            if cache_key in _line_cache:
                value = _line_cache[cache_key]
            else:
                try:
                    value = ColouredText(line, self._parser, colour=last_colour)
                except IndexError:
                    # TODO: Create an issue on github for this
                    continue
                _line_cache[cache_key] = value

            # value = ColouredText(line, self._parser, colour=last_colour)
            new_lines.append(value)
            last_colour = value.last_colour

            # Release chunks of values at a time, so the UI has something to update with
            if len(new_lines) > 100:
                self._processed_data_queue.put(new_lines)
                new_lines = []

        # Put this in a queue to be picked up by the main thread, in self.update
        self._processed_data_queue.put(new_lines)

    @property
    def value(self):
        self._value = [ColouredText("", self._parser, colour=None)]
        return self._value

    @value.setter
    def value(self, value):
        """This value should never be used. It's been replaced by _add_stream"""
        pass



import json
from pathlib import Path
from typing import Dict, Iterator

from groklog.filter_manager import exceptions
from groklog.process_node import GenericProcessIO, ProcessNode, ShellProcessIO

ROOT_FILTER_NAME = "Shell"
"""The default name for the root shell process"""


class FilterManager:
    """
    This class handles the storing, saving, loading, and creation of ProcessNode objects
    """

    # Set constants for how serialization. The UI also uses these for labels.
    FILTER_PARENT = "parent_process"
    FILTER_NAME = "name"
    FILTER_COMMAND = "command"
    FILTER_CHILDREN = "children"

    def __init__(self, shell: ShellProcessIO):
        """
        :param shell: The shell, which will be the 'root' process for input
        """
        self.selected_filter = shell

        self._filters: Dict[str, Filter] = {}
        """A dictionary of Filter.name: Filter"""

        # Register the "root" filter
        self._filters[ROOT_FILTER_NAME] = shell

    def __iter__(self) -> Iterator[ProcessNode]:
        for filter in self._filters.values():
            yield filter

    @property
    def root_filter(self) -> ShellProcessIO:
        return self.get_filter(ROOT_FILTER_NAME)

    def get_filter(self, filter_name: str) -> ProcessNode:
        try:
            return self._filters[filter_name]
        except KeyError:
            raise exceptions.FilterNotFoundError(
                f"Could not find a filter of name {filter_name}"
            )

    def create_filter(
        self,
        name: str,
        command: str,
        parent: ProcessNode,
    ) -> ProcessNode:
        """Create and register a new filter.
        :param name: The name of the filter
        :param command: The shell command to run
        :param parent: The children process to feed results into the new filter
        :return: The new filter
        """

        if name in self._filters:
            raise exceptions.DuplicateFilterError(
                f"A filter with name '{name}' already exists!"
            )

        # Create and register the filter
        filter = GenericProcessIO(
            name=name,
            command=command,
        )
        parent.add_child(filter)
        self._filters[name] = filter
        return filter

    def save_profile(self, profile_path: Path):
        def serialize_process_node(process_node: ProcessNode):
            return {
                self.FILTER_NAME: process_node.name,
                self.FILTER_COMMAND: process_node.command,
                self.FILTER_CHILDREN: [
                    serialize_process_node(c) for c in process_node.children
                ],
            }

        profile_path.parent.mkdir(parents=True, exist_ok=True)
        with profile_path.open("w") as file:
            json.dump(serialize_process_node(self.root_filter), file)

    def load_profile(self, profile_path: Path):
        def deserialize_process_node(node_info, parent=None):
            if parent is None:
                node = self.root_filter
            else:
                node = self.create_filter(
                    name=node_info[self.FILTER_NAME],
                    command=node_info[self.FILTER_COMMAND],
                    parent=parent,
                )

            for child_info in node_info[self.FILTER_CHILDREN]:
                deserialize_process_node(child_info, parent=node)
            return node

        with profile_path.open("r") as file:
            profile_json = json.load(file)
        deserialize_process_node(profile_json)

    def close(self):
        self.root_filter.close()



from argparse import ArgumentParser
from pathlib import Path

import appdirs

long_description = """
Welcome to GrokLog!

GrokLog is a tool for creating "process trees", that is, a root process (an ordinary shell)
has its output piped into the stdin of other processes. Those other processes can in 
turn have _their_ stdout piped into even more processes. Then, groklog lets you view the 
stdout of every process in the tree. 
 
The benefit of all of this is that you can filter logs or other streams using the 
various unix tools you are already familiar with, and do so in a much more sophisticated 
way. 

On top of that, building profiles is quick and easy, so after spending some time 
configuring, everything is saved to a profile and the next time you run groklog you can 
jump right in where you left off!
"""


def parse_args():
    parser = ArgumentParser(description=long_description)

    default_save_path = Path(
        appdirs.user_config_dir(appname="groklog", appauthor="Alex Thiel")
    )

    parser.add_argument(
        "--profile-directory",
        default=default_save_path,
        help="Override the default directory to load profiles from.",
    )

    parser.add_argument(
        "profile",
        type=str,
        nargs="?",
        default="default",
        help="Load a custom profile. If nothing is entered, the default profile "
        f"at {default_save_path / 'default.json'} will be used.",
    )

    return parser.parse_args()



from abc import ABC, abstractmethod
from enum import Enum, auto
from queue import Queue
from threading import RLock
from typing import Callable, List

from pubsus import DuplicateSubscriberError, PubSubMixin


class ProcessNode(ABC, PubSubMixin):
    _READ_MAX_BYTES = 102400

    class Topic(Enum):
        # TODO: Add a topic that publishes only when full, completed lines are ready.

        STRING_DATA_STREAM = auto()
        """The process's stdout data, but converted to a utf-8 encoded string"""

        BYTES_DATA_STREAM = auto()
        """The process's stdout data as raw bytes"""

    def __init__(self, name: str, command: str):
        """
        :param name: An arbitrary unique title for this process
        :param command: The command being run in the process
        """
        super().__init__()
        self.name = name
        self.command = command
        self.children: List[ProcessNode] = []

        self._new_subscribers: Queue[Tuple[ProcessNode.Topic, Callable]] = Queue()
        """This queue contains incoming subscribers that have requested to have 
        the full history applied. This is a special case, because the callback must be
        called once with the full history. This is done in the background thread, so that
        subscribing isn't a blocking operation. """

        self._history_lock = RLock()
        self._bytes_history: bytes = b""
        self._string_history: str = ""

    def __repr__(self):
        return f"{self.__class__.__qualname__}(name='{self.name}', command='{self.command}')"

    def add_child(self, process_node: "ProcessNode"):
        """Adds and subscribes the child"""
        self.subscribe_with_history(
            ProcessNode.Topic.BYTES_DATA_STREAM, process_node.write, blocking=False
        )
        self.children.append(process_node)

    def _record_and_publish(self, data_bytes: bytes):
        """Record the data to the history and publish the bytes and string
        variants of the data."""
        # Scrub all control characters out of the string so they don't get passed on to
        # child process and kill them too

        data_string: str = data_bytes.decode("utf8", "replace")

        with self._history_lock:
            self._bytes_history += data_bytes
            self._string_history += data_string

            self.publish(self.Topic.STRING_DATA_STREAM, data_string)
            self.publish(self.Topic.BYTES_DATA_STREAM, data_bytes)

    def subscribe_with_history(
        self, topic: "ProcessNode.Topic", subscriber: Callable, *, blocking
    ):
        """This function will subscribe a subscriber to the full history that has ever
        been received by this node. The callback will occur in the background thread.
        """
        if blocking:
            self._onboard_subscriber(topic, subscriber)
        else:
            self._new_subscribers.put((topic, subscriber))

    def _onboard_new_subscribers(self):
        """Onboard any subscribers who wish to have the full history before adding more
        history.
        """

        with self._history_lock:
            while self._running and self._new_subscribers.qsize():
                self._onboard_subscriber(*self._new_subscribers.get_nowait())
                self._new_subscribers.task_done()

    def _onboard_subscriber(self, topic, subscriber):
        if self.is_subscribed(topic, subscriber):
            raise DuplicateSubscriberError("This topic/subscriber already exists!")
        with self._history_lock:
            # If any history has been written, pass it along
            if len(self._string_history):
                if topic is self.Topic.STRING_DATA_STREAM:
                    subscriber(self._string_history)
                elif topic is self.Topic.BYTES_DATA_STREAM:
                    subscriber(self._bytes_history)
                else:
                    raise ValueError("Invalid topic")
            self.subscribe(topic, subscriber)

    @abstractmethod
    def write(self, val: bytes):
        pass

    def close(self, timeout=None):
        """Close all child processes and threads"""
        self._running = False
        self._process.kill()
        self._extraction_thread.join(timeout)
        self._process.wait(timeout)

        for child in self.children:
            child.close(timeout=timeout)



#!/usr/bin/env python3
import curses
import fcntl
import struct
import termios
from queue import Queue

from asciimatics.event import KeyboardEvent
from asciimatics.parsers import AnsiTerminalParser, Parser
from asciimatics.screen import Canvas, Screen
from asciimatics.widgets import Widget

from groklog.process_node import ShellProcessIO


class Terminal(Widget):
    """
    Widget to handle ansi terminals running a bash shell.
    """

    def __init__(
        self,
        name: str,
        shell: ShellProcessIO,
        height: int,
        show_cursor: bool = True,
    ):
        super().__init__(name)
        self._required_height = height
        self._parser = AnsiTerminalParser()
        self._canvas = None
        self._current_colours = None
        self._cursor_x, self._cursor_y = 0, 0
        self._show_cursor = show_cursor

        # Supported key mappings
        self._map = {}
        for k, v in [
            (Screen.KEY_LEFT, "kcub1"),
            (Screen.KEY_RIGHT, "kcuf1"),
            (Screen.KEY_UP, "kcuu1"),
            (Screen.KEY_DOWN, "kcud1"),
            (Screen.KEY_PAGE_UP, "kpp"),
            (Screen.KEY_PAGE_DOWN, "knp"),
            (Screen.KEY_HOME, "khome"),
            (Screen.KEY_END, "kend"),
            (Screen.KEY_DELETE, "kdch1"),
            (Screen.KEY_BACK, "kbs"),
        ]:
            self._map[k] = curses.tigetstr(v)
        self._map[Screen.KEY_TAB] = "\t".encode()

        # Subscribe to shell data
        self._data_queue = Queue()
        self._shell = shell

    def update(self, frame_no):
        """Draw the current terminal content to screen."""
        # Push current terminal output to screen.
        self._canvas.refresh()

        # Drain the shell queue of any data that built up between frames
        # TODO: Pull only processed results from the stream
        full_stream = ""
        while self._data_queue.qsize():
            full_stream += self._data_queue.get_nowait()
        if len(full_stream):
            self._add_stream(full_stream)

        # Draw cursor if needed.
        if frame_no % 20 < 10 and self._show_cursor:
            origin = self._canvas.origin
            x = self._cursor_x + origin[0]
            y = self._cursor_y + origin[1] - self._canvas.start_line
            details = self._canvas.get_from(self._cursor_x, self._cursor_y)
            if details:
                char, colour, attr, bg = details
                attr |= Screen.A_REVERSE
                self._frame.canvas.print_at(chr(char), x, y, colour, attr, bg)

    def set_layout(self, x, y, offset, w, h):
        """
        Resize the widget (and underlying TTY) to the required size.
        """
        super().set_layout(x, y, offset, w, h)
        self._canvas = Canvas(self._frame.canvas, h, w, x=x, y=y)
        winsize = struct.pack("HHHH", h, w, 0, 0)
        fcntl.ioctl(self._shell._slave, termios.TIOCSWINSZ, winsize)

    def process_event(self, event):
        """
        Pass any recognised input on to the TTY.
        """
        if isinstance(event, KeyboardEvent):
            if event.key_code > 0:
                val = chr(event.key_code).encode()
            elif event.key_code in self._map:
                val = self._map[event.key_code]
            self._shell.write(val)
            return
        return event

    def _add_stream(self, value: str):
        """
        Process any output from the TTY. The value is split into lines, and only
        the last X lines will get rendered. This makes it possible for the
        shell to dump huge amounts of data, but only the lines that fit on the
        height of the terminal will ever be rendered.
        """
        canvas_height, _ = self._canvas.dimensions
        lines = value.split("\n")[-canvas_height:]

        for i, line in enumerate(lines):
            self._parser.reset(line, self._current_colours)
            for offset, command, params in self._parser.parse():
                if command == Parser.DISPLAY_TEXT:
                    # Just display the text...  allowing for line wrapping.
                    if self._cursor_x + len(params) > self._w:
                        part_1 = params[: self._w - self._cursor_x]
                        part_2 = params[self._w - self._cursor_x :]
                        self._print_at(part_1, self._cursor_x, self._cursor_y)
                        self._print_at(part_2, 0, self._cursor_y + 1)
                        self._cursor_x = len(part_2)
                        self._cursor_y += 1
                        if self._cursor_y - self._canvas.start_line >= self._h:
                            self._canvas.scroll()
                    else:
                        self._print_at(params, self._cursor_x, self._cursor_y)
                        self._cursor_x += len(params)
                elif command == Parser.CHANGE_COLOURS:
                    # Change current text colours.
                    self._current_colours = params
                elif command == Parser.NEXT_TAB:
                    # Move to next tab stop - hard-coded to default of 8 characters.
                    self._cursor_x = (self._cursor_x // 8) * 8 + 8
                elif command == Parser.MOVE_RELATIVE:
                    # Move cursor relative to current position.
                    self._cursor_x += params[0]
                    self._cursor_y += params[1]
                    if self._cursor_y < self._canvas.start_line:
                        self._canvas.scroll(self._cursor_y - self._canvas.start_line)
                elif command == Parser.MOVE_ABSOLUTE:
                    # Move cursor relative to specified absolute position.
                    if params[0] is not None:
                        self._cursor_x = params[0]
                    if params[1] is not None:
                        self._cursor_y = params[1] + self._canvas.start_line
                elif command == Parser.DELETE_LINE:
                    # Delete some/all of the current line.
                    if params == 0:
                        self._print_at(
                            " " * (self._w - self._cursor_x),
                            self._cursor_x,
                            self._cursor_y,
                        )
                    elif params == 1:
                        self._print_at(" " * self._cursor_x, 0, self._cursor_y)
                    elif params == 2:
                        self._print_at(" " * self._w, 0, self._cursor_y)
                elif command == Parser.DELETE_CHARS:
                    # Delete n characters under the cursor.
                    for x in range(self._cursor_x, self._w):
                        if x + params < self._w:
                            cell = self._canvas.get_from(x + params, self._cursor_y)
                        else:
                            cell = (
                                ord(" "),
                                self._current_colours[0],
                                self._current_colours[1],
                                self._current_colours[2],
                            )
                        self._canvas.print_at(
                            chr(cell[0]),
                            x,
                            self._cursor_y,
                            colour=cell[1],
                            attr=cell[2],
                            bg=cell[3],
                        )
                elif command == Parser.SHOW_CURSOR:
                    # Show/hide the cursor.
                    self._show_cursor = params
                elif command == Parser.CLEAR_SCREEN:
                    # Clear the screen.
                    self._canvas.clear_buffer(
                        self._current_colours[0],
                        self._current_colours[1],
                        self._current_colours[2],
                    )
            # Move to next line, scrolling buffer as needed.
            if i != len(lines) - 1:
                self._cursor_x = 0
                self._cursor_y += 1
                if self._cursor_y - self._canvas.start_line >= self._h:
                    self._canvas.scroll()

    def _print_at(self, text, x, y):
        """
        Helper function to simplify use of the canvas.
        """
        self._canvas.print_at(
            text,
            x,
            y,
            colour=self._current_colours[0],
            attr=self._current_colours[1],
            bg=self._current_colours[2],
        )

    def reset(self):
        """
        Reset the widget to a blank screen.
        """
        self._canvas = Canvas(
            self._frame.canvas, self._h, self._w, x=self._x, y=self._y
        )
        self._cursor_x, self._cursor_y = 0, 0
        self._current_colours = (
            Screen.COLOUR_WHITE,
            Screen.A_NORMAL,
            Screen.COLOUR_BLACK,
        )

        # Unsubscribe and resubscribe to the stream to get back the full history
        if self._shell.is_subscribed(
            ShellProcessIO.Topic.STRING_DATA_STREAM, self._data_queue.put
        ):
            self._shell.unsubscribe(
                ShellProcessIO.Topic.STRING_DATA_STREAM, self._data_queue.put
            )

        # TODO: Is there a cleaner way of doing this
        self._shell.subscribe_with_history(
            ShellProcessIO.Topic.STRING_DATA_STREAM, self._data_queue.put, blocking=True
        )

    def required_height(self, offset, width):
        """
        Required height for the terminal.
        """
        return self._required_height

    @property
    def frame_update_count(self):
        """
        Frame update rate required.
        """
        # Force refresh for cursor.
        return 1

    @property
    def value(self):
        """
        Terminal value - not needed for demo.
        """
        return

    @value.setter
    def value(self, _):
        return



from asciimatics import widgets
from asciimatics.screen import Screen

GROKLOG_THEME = {
    "background": (Screen.COLOUR_WHITE, Screen.A_NORMAL, Screen.COLOUR_BLACK),
    "shadow": (Screen.COLOUR_WHITE, None, Screen.COLOUR_WHITE),
    "disabled": (Screen.COLOUR_WHITE, Screen.A_BOLD, Screen.COLOUR_BLACK),
    "invalid": (Screen.COLOUR_WHITE, Screen.A_BOLD, Screen.COLOUR_YELLOW),
    "label": (Screen.COLOUR_BLUE, Screen.A_BOLD, Screen.COLOUR_BLACK),
    "borders": (Screen.COLOUR_WHITE, Screen.A_BOLD, Screen.COLOUR_BLACK),
    "scroll": (Screen.COLOUR_CYAN, Screen.A_NORMAL, Screen.COLOUR_BLACK),
    "title": (Screen.COLOUR_WHITE, Screen.A_BOLD, Screen.COLOUR_BLACK),
    "edit_text": (Screen.COLOUR_WHITE, Screen.A_NORMAL, Screen.COLOUR_BLACK),
    "focus_edit_text": (Screen.COLOUR_WHITE, Screen.A_BOLD, Screen.COLOUR_CYAN),
    "readonly": (Screen.COLOUR_WHITE, Screen.A_BOLD, Screen.COLOUR_BLACK),
    "focus_readonly": (Screen.COLOUR_CYAN, Screen.A_BOLD, Screen.COLOUR_CYAN),
    "button": (Screen.COLOUR_WHITE, Screen.A_NORMAL, Screen.COLOUR_BLACK),
    "focus_button": (Screen.COLOUR_WHITE, Screen.A_BOLD, Screen.COLOUR_CYAN),
    "control": (Screen.COLOUR_YELLOW, Screen.A_NORMAL, Screen.COLOUR_BLACK),
    "selected_control": (Screen.COLOUR_YELLOW, Screen.A_BOLD, Screen.COLOUR_BLACK),
    "focus_control": (Screen.COLOUR_YELLOW, Screen.A_NORMAL, Screen.COLOUR_BLACK),
    "selected_focus_control": (
        Screen.COLOUR_YELLOW,
        Screen.A_BOLD,
        Screen.COLOUR_CYAN,
    ),
    "field": (Screen.COLOUR_WHITE, Screen.A_NORMAL, Screen.COLOUR_BLACK),
    "selected_field": (Screen.COLOUR_YELLOW, Screen.A_BOLD, Screen.COLOUR_BLACK),
    "focus_field": (Screen.COLOUR_WHITE, Screen.A_NORMAL, Screen.COLOUR_BLACK),
    "selected_focus_field": (
        Screen.COLOUR_WHITE,
        Screen.A_BOLD,
        Screen.COLOUR_CYAN,
    ),
    # This is a custom color for the FilterViewer widget
    "filter_viewer": (Screen.COLOUR_DEFAULT, Screen.A_BOLD, Screen.COLOUR_BLACK),
}

warning = widgets.utilities.THEMES["warning"]
warning["shadow"] = (Screen.COLOUR_BLACK, None, Screen.COLOUR_WHITE)


def set_theme(frame):
    widgets.utilities.THEMES["groklog"] = GROKLOG_THEME
    frame.set_theme("groklog")



import fcntl
import itertools
import os
import re
import shlex
import subprocess
import unicodedata
from threading import RLock, Thread
from time import sleep

from .base import ProcessNode


class GenericProcessIO(ProcessNode):
    def __init__(self, name: str, command: str):
        super().__init__(name=name, command=command)

        self._process = subprocess.Popen(
            shlex.split(self.command),
            shell=False,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )

        # Set stdout to be nonblocking
        fd = self._process.stdout.fileno()
        fl = fcntl.fcntl(fd, fcntl.F_GETFL)
        fcntl.fcntl(fd, fcntl.F_SETFL, fl | os.O_NONBLOCK)

        # Lock while processes are piping data in
        self._input_lock = RLock()

        self._running = True
        self._extraction_thread = Thread(
            name=f"Thread({self})", daemon=True, target=self._background
        )
        self._extraction_thread.start()

    def write(self, data: bytes):
        """Input data from an upstream process"""
        if not self._running:
            return

        with self._input_lock:
            self._process.stdin.write(data)
            self._process.stdin.flush()

    def _background(self):
        # control_chars = bytes(''.join(map(chr, itertools.chain(range(0x00,0x20), range(0x7f,0xa0)))).encode("utf-8"))
        # control_char_re = re.compile(b'[%s]' % re.escape(control_chars))

        control_chars = bytes(b"".join([b"0x03", b"0x04", b"0x018", b"0x00"]))
        control_char_re = re.compile(b"[%s]" % re.escape(b"\x03"))

        while self._running:
            self._onboard_new_subscribers()

            # Read in a non-blocking fashion up to a certain number of characters.
            data_bytes = self._process.stdout.read1(self._READ_MAX_BYTES)

            if len(data_bytes) == 0:
                # Prevent a busyloop where there's no data in stdout.
                sleep(0.1)
                continue

            # data_bytes = data_bytes.replace(bytes(0x03), b'')
            # data_bytes = data_bytes.replace(bytes(0x04), b'')
            # data_bytes = data_bytes.replace(bytes(0x018), b'')
            # data_bytes = control_char_re.sub(b'', data_bytes)
            # data_bytes = bytes(c for c in data_bytes if c == 3)
            data_bytes = data_bytes.replace(b"\x03", b"")
            self._record_and_publish(data_bytes)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['groklog',
 'groklog.filter_manager',
 'groklog.process_node',
 'groklog.ui',
 'groklog.ui.scenes']

package_data = \
{'': ['*']}

install_requires = \
['asciimatics>=1.13.0,<2.0.0',
 'pre-commit>=2.13.0,<3.0.0',
 'pubsus>=0.1.1,<0.2.0']

entry_points = \
{'console_scripts': ['groklog = groklog:main']}

setup_kwargs = {
    'name': 'groklog',
    'version': '0.1.0',
    'description': 'A tool for filtering logs quickly and easily',
    'long_description': None,
    'author': 'Alex Thiel',
    'author_email': 'apocthiel@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import os
import sys

from asciimatics.exceptions import ResizeScreenError
from asciimatics.scene import Scene
from asciimatics.screen import Screen

from groklog.args import parse_args
from groklog.filter_manager import FilterManager
from groklog.process_node import ShellProcessIO
from groklog.ui.scenes import FilterCreator, GrokLog, scene_names

from .version import __version__


def main():
    # In some systems, curses takes a while to pass Escape key input. More info here:
    # https://github.com/peterbrittain/asciimatics/issues/232
    os.environ.setdefault("ESCDELAY", "0")
    args = parse_args()

    filter_manager = FilterManager(shell=ShellProcessIO())

    # Load configuration
    save_path = args.profile_directory / (args.profile + ".json")
    if save_path.is_file():
        filter_manager.load_profile(save_path)

    def groklog(screen: Screen, scene):
        scenes = [
            Scene(
                [GrokLog(screen, filter_manager=filter_manager)],
                duration=-1,
                name=scene_names.SHELL_VIEW,
            ),
            Scene(
                [
                    FilterCreator(
                        screen, filter_manager=filter_manager, profile_path=save_path
                    )
                ],
                duration=-1,
                name=scene_names.FILTER_CREATOR_SCENE,
            ),
        ]

        screen.play(scenes, stop_on_resize=True, start_scene=scene, allow_int=True)

    last_scene = None
    while True:
        try:
            Screen.wrapper(func=groklog, catch_interrupt=True, arguments=[last_scene])
            print("Thank you for using GrokLog!")
            sys.exit(0)
        except ResizeScreenError as e:
            last_scene = e.scene


if __name__ == "__main__":
    main()



from functools import partial

from asciimatics import widgets
from asciimatics.event import KeyboardEvent
from asciimatics.screen import Screen
from asciimatics.widgets import Layout

from groklog.filter_manager import FilterManager
from groklog.process_node import ShellProcessIO
from groklog.ui.filter_viewer import FilterViewer
from groklog.ui.terminal import Terminal

from . import scene_names
from .base_app import BaseApp


class GrokLog(BaseApp):
    def __init__(self, screen, filter_manager: FilterManager):
        super().__init__(
            screen,
            screen.height,
            screen.width,
            has_border=True,
            can_scroll=False,
            name="GrokLog",
            title="GrokLog",
        )
        self.filter_manager = filter_manager

        # Register all of the filter widgets
        self._filter_widgets = {}
        for filter in self.filter_manager:
            self._maybe_register_filter(filter)

        self.central_layout = Layout([100], fill_frame=True)
        self.add_layout(self.central_layout)
        self.view_filter(self.filter_manager.selected_filter)

        # Create the Tab Layout and buttons for it
        self.tab_layout = Layout([1, 0, 1, 1, 1, 1, 1])
        self.add_layout(self.tab_layout)
        self.create_tab_buttons()

        self.fix()

    def reset(self):
        # After coming back from the AddFilter call, reacreate the tab buttons to fill
        # in any missing tabs.
        for filter in self.filter_manager:
            self._maybe_register_filter(filter)

        self.create_tab_buttons()
        return super().reset()

    def _maybe_register_filter(self, filter):
        """Create a widget for this filter and save it under self._filter_widgets"""
        if filter in self._filter_widgets:
            # This filter is already registered.
            return

        if isinstance(filter, ShellProcessIO):
            widget = Terminal(
                name="term",
                shell=filter,
                height=widgets.Widget.FILL_COLUMN,
            )
        else:
            widget = FilterViewer(filter=filter, height=widgets.Widget.FILL_COLUMN)

        self._filter_widgets[filter] = widget

    def view_filter(self, filter):
        """Change the actively shown central widget"""
        if self.scene is not None:
            self.display_toast(f"Viewing {filter.name}: '{filter.command}'")

        self.filter_manager.selected_filter = filter

        # Replace the central layout widget
        new_widget = self._filter_widgets[filter]
        self.central_layout.clear_widgets()
        self.central_layout.add_widget(new_widget)
        self.central_layout.add_widget(widgets.Divider())

        if isinstance(new_widget, Terminal):
            # The terminal has a... hard time keeping stuff on the screen. This forces
            # the terminal to re-subscribe and refresh the screen.
            new_widget.reset()

        # This seems to put the widget into the update() loop
        self.central_layout.focus(force_widget=new_widget)
        self.screen.force_update(full_refresh=True)
        self.fix()

    def create_tab_buttons(self):
        """Create all of the tab buttons again"""
        self.tab_layout.clear_widgets()
        self.tab_layout.add_widget(
            widgets.Button(
                text="Add Filter",
                on_click=partial(self.change_scene, scene_names.FILTER_CREATOR_SCENE),
            ),
            column=0,
        )
        self.tab_layout.add_widget(widgets.VerticalDivider(), column=1)

        for column, filter in enumerate(self.filter_manager, 2):
            self.tab_layout.add_widget(
                widgets.Button(
                    text=filter.name,
                    on_click=lambda filter=filter: self.view_filter(filter),
                ),
                column=column,
            )

        self.fix()

    def process_event(self, event):
        if isinstance(event, KeyboardEvent):
            if event.key_code in [Screen.ctrl("c")]:
                # Catch Ctrl+C and pass it on to the sub shell
                self.display_toast("Press Escape to close GrokLog!")
                if (
                    self.filter_manager.selected_filter
                    is self.filter_manager.root_filter
                ):
                    self.filter_manager.root_filter.send_sigint()
                return

        return super().process_event(event)
